try:
    from pytz import UTC, _UTC # we import _UTC so that pickles made by the
    # fallback code below can still be reinstantiated if pytz is added in.
except ImportError:
    import datetime
    class UTC(datetime.tzinfo):
        """UTC
        
        Identical to the reference UTC implementation given in Python docs except
        that it unpickles using the single module global instance defined beneath
        this class declaration.
    
        Also contains extra attributes and methods to match other pytz tzinfo
        instances.
        """
        zone = "UTC"
    
        def utcoffset(self, dt):
            return ZERO
    
        def tzname(self, dt):
            return "UTC"
    
        def dst(self, dt):
            return ZERO
        
        def __reduce__(self):
            return _UTC, ()
    
        def localize(self, dt, is_dst=False):
            '''Convert naive time to local time'''
            if dt.tzinfo is not None:
                raise ValueError, 'Not naive datetime (tzinfo is already set)'
            return dt.replace(tzinfo=self)
    
        def normalize(self, dt, is_dst=False):
            '''Correct the timezone information on the given datetime'''
            if dt.tzinfo is None:
                raise ValueError, 'Naive time - no tzinfo set'
            return dt.replace(tzinfo=self)
    
        def __repr__(self):
            return "<UTC>"
    
        def __str__(self):
            return "UTC"
    
    
    UTC = utc = UTC() # UTC is a singleton
    
    
    def _UTC():
        """Factory function for utc unpickling.
        
        Makes sure that unpickling a utc instance always returns the same 
        module global.
        
        These examples belong in the UTC class above, but it is obscured; or in
        the README.txt, but we are not depending on Python 2.4 so integrating
        the README.txt examples with the unit tests is not trivial.
        
        >>> import datetime, pickle
        >>> dt = datetime.datetime(2005, 3, 1, 14, 13, 21, tzinfo=utc)
        >>> naive = dt.replace(tzinfo=None)
        >>> p = pickle.dumps(dt, 1)
        >>> naive_p = pickle.dumps(naive, 1)
        >>> len(p), len(naive_p), len(p) - len(naive_p)
        (60, 43, 17)
        >>> new = pickle.loads(p)
        >>> new == dt
        True
        >>> new is dt
        False
        >>> new.tzinfo is dt.tzinfo
        True
        >>> utc is UTC is timezone('UTC')
        True
        >>> utc is timezone('GMT')
        False
        """
        return utc
    _UTC.__safe_for_unpickling__ = True



##############################################################################
#
# Copyright (c) 2006-2008 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import datetime
import BTrees
import zope.bforest.bforest
import zope.bforest.utils

def mutating(name):
    def mutate(self, *args, **kwargs):
        if (datetime.datetime.now(zope.bforest.utils.UTC) -
            self.last_rotation) >= self._inner_period:
            self.rotateBucket()
        return getattr(super(Abstract, self), name)(*args, **kwargs)
    return mutate

class Abstract(zope.bforest.bforest.AbstractBForest):
    def __init__(self, period, d=None, count=2):
        super(Abstract, self).__init__(d, count)
        self.period = period
        self.last_rotation = datetime.datetime.now(zope.bforest.utils.UTC)

    _inner_period = _period = None
    def period(self, value):
        self._period = value
        self._inner_period = self._period / (len(self.buckets) - 1)
    period = property(
        lambda self: self._period, period)
    
    def copy(self):
        # this makes an exact copy, including the individual state of each 
        # bucket.  If you want a dict, cast it to a dict, or if you want
        # another one of these but with all of the keys in the first bucket,
        # call obj.__class__(obj)
        copy = self.__class__(self.period, count=len(self.buckets))
        for i in range(len(self.buckets)):
            copy.buckets[i].update(self.buckets[i])
        return copy

    def rotateBucket(self):
        super(Abstract, self).rotateBucket()
        self.last_rotation = datetime.datetime.now(zope.bforest.utils.UTC)

    __setitem__ = mutating('__setitem__')
    __delitem__ = mutating('__delitem__')
    pop = mutating('pop')
    popitem = mutating('popitem')
    update = mutating('update')

class IOBForest(Abstract):
    _treemodule = BTrees.family32.IO

class OIBForest(Abstract):
    _treemodule = BTrees.family32.OI

class IIBForest(Abstract):
    _treemodule = BTrees.family32.II

class LOBForest(Abstract):
    _treemodule = BTrees.family64.IO

class OLBForest(Abstract):
    _treemodule = BTrees.family64.OI

class LLBForest(Abstract):
    _treemodule = BTrees.family64.II

class OOBForest(Abstract):
    _treemodule = BTrees.family32.OO



##############################################################################
#
# Copyright (c) 2006-2008 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import os
from setuptools import setup, find_packages

# generic helpers primarily for the long_description
try:
    import docutils
except ImportError:
    import warnings
    def validateReST(text):
        return ''
else:
    import docutils.utils
    import docutils.parsers.rst
    import StringIO
    def validateReST(text):
        doc = docutils.utils.new_document('validator')
        # our desired settings
        doc.reporter.halt_level = 5
        doc.reporter.report_level = 1
        stream = doc.reporter.stream = StringIO.StringIO()
        # docutils buglets (?)
        doc.settings.tab_width = 2
        doc.settings.pep_references = doc.settings.rfc_references = False
        doc.settings.trim_footnote_reference_space = None
        # and we're off...
        parser = docutils.parsers.rst.Parser()
        parser.parse(text, doc)
        return stream.getvalue()

def text(*args, **kwargs):
    # note: distutils explicitly disallows unicode for setup values :-/
    # http://docs.python.org/dist/meta-data.html
    tmp = []
    for a in args:
        if a.endswith('.txt'):
            f = open(os.path.join(*a.split('/')))
            tmp.append(f.read())
            f.close()
            tmp.append('\n\n')
        else:
            tmp.append(a)
    if len(tmp) == 1:
        res = tmp[0]
    else:
        res = ''.join(tmp)
    out = kwargs.get('out')
    if out is True:
        out = 'TEST_THIS_REST_BEFORE_REGISTERING.txt'
    if out:
        f = open(out, 'w')
        f.write(res)
        f.close()
        report = validateReST(res)
        if report:
            print report
            raise ValueError('ReST validation error')
    return res
# end helpers; below this line should be code custom to this package

setup(
    name="zope.bforest",
    version="1.2",
    license="ZPL 2.1",
    author='Gary Poster',
    author_email='gary@zope.com',

    namespace_packages=["zope"],
    packages=find_packages('src'),
    package_dir={'':'src'},
    include_package_data=True,
    install_requires=[
        "zope.interface",
        "ZODB3",
        "setuptools",
        ],
    tests_require=["zope.testing"],
    description=open('README.txt').read(),
    long_description=text(
        "src/zope/bforest/bforest.txt",
        "\n\n",
        'CHANGES.txt',
        out=True),
    keywords="zope zope3",
    zip_safe=False
    )



##############################################################################
#
# Copyright (c) 2004 Zope Corporation and Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""objects with dict API comprised of multiple btrees.
"""

import persistent
import persistent.list
import BTrees

class AbstractBForest(persistent.Persistent):

    _treemodule = None # override
    _marker = object()
    
    def __init__(self, d=None, count=2):
        if count < 0:
            raise ValueError("count must be 0 or greater")
        if count == 0:
            if d is not None:
                raise ValueError(
                    "cannot set initial values without a bucket")
            l = persistent.list.PersistentList()
        else:
            Tree = self._treemodule.BTree
            if d is not None:
                first = Tree(d)
            else:
                first = Tree()
            l = [Tree() for i in range(count - 1)]
            l.insert(0, first)
            l = persistent.list.PersistentList(l)
        self.buckets = l
    
    def __getitem__(self, key):
        m = self._marker
        res = self.get(key, m)
        if res is m:
            raise KeyError(key)
        return res
    
    def __setitem__(self, key, value):
        self.buckets[0][key] = value
    
    def get(self, key, default=None):
        m = self._marker
        for b in self.buckets:
            res = b.get(key, m)
            if res is not m:
                return res
        return default
    
    def __delitem__(self, key):
        found = False
        for b in self.buckets:
            # we delete all, in case one bucket is masking an old value
            try:
                del b[key]
            except KeyError:
                pass
            else:
                found = True
        if not found:
            raise KeyError(key)
    
    def update(self, d):
        self.buckets[0].update(d)
    
    def keys(self):
        buckets = self.buckets
        if len(buckets) == 1:
            res = buckets[0].keys()
        elif getattr(self._treemodule, 'multiunion', None) is not None:
            return self._treemodule.multiunion(buckets)
        else:
            union = self._treemodule.union
            res = union(buckets[0], buckets[1])
            for b in buckets[2:]:
                res = union(res, b)
        return res
    
    def tree(self):
        # convert to a tree; do as much in C as possible.
        buckets = self.buckets
        res = self._treemodule.BTree(buckets[-1])
        for b in buckets[-2::-1]:
            res.update(b)
        return res
    
    def items(self):
        return self.tree().items()

    def values(self):
        return self.tree().values()

    def maxKey(self, key=None):
        res = m = self._marker
        if key is None:
            args = ()
        else:
            args = (key,)
        for b in self.buckets:
            try:
                v = b.maxKey(*args)
            except ValueError:
                pass
            else:
                if res is m or v > res:
                    res = v
        if res is m:
            raise ValueError('no key')
        return res

    def minKey(self, key=None):
        res = m = self._marker
        if key is None:
            args = ()
        else:
            args = (key,)
        for b in self.buckets:
            try:
                v = b.minKey(*args)
            except ValueError:
                pass
            else:
                if res is m or v < res:
                    res = v
        if res is m:
            raise ValueError('no key')
        return res

    def iteritems(self, min=None, max=None, excludemin=False, excludemax=False):
        # this approach might be slower than simply iterating over self.keys(),
        # but it has a chance of waking up fewer objects in the ZODB, if the
        # iteration stops early.
        sources = []
        for b in self.buckets:
            i = iter(b.items(min, max, excludemin, excludemax))
            try:
                first = i.next()
            except StopIteration:
                pass
            else:
                sources.append([first, i])
        scratch = self._treemodule.Bucket()
        while sources:
            for ((k, v), i) in sources:
                scratch.setdefault(k, v)
            k = scratch.minKey()
            yield k, scratch[k]
            for i in range(len(sources)-1, -1, -1):
                source = sources[i]
                if source[0][0] == k:
                    try:
                        source[0] = source[1].next()
                    except StopIteration:
                        del sources[i]
            scratch.clear()
    
    def iterkeys(self,
                 min=None, max=None, excludemin=False, excludemax=False):
        return (k for k, v in self.iteritems(min, max, excludemin, excludemax))
    
    __iter__ = iterkeys
    
    def itervalues(self,
                   min=None, max=None, excludemin=False, excludemax=False):
        return (v for k, v in self.iteritems(min, max, excludemin, excludemax))
    
    def __eq__(self, other):
        # we will declare contents and its ordering to define equality.  More
        # restrictive definitions can be wrapped around this one.
        if getattr(other, 'iteritems', None) is None:
            return False
        si = self.iteritems()
        so = other.iteritems()
        while 1:
            try:
                k, v = si.next()
            except StopIteration:
                try:
                    so.next()
                except StopIteration:
                    return True
                return False
            try:
                ok, ov = so.next()
            except StopIteration:
                return False
            else:
                if ok != k or ov != v:
                    return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __gt__(self, other):
        # TODO blech, change this
        if not isinstance(other, dict):
            try:
                other = dict(other)
            except TypeError:
                return id(self) > id(other)
        return dict(self) > other

    def __lt__(self, other):
        # TODO blech, change this
        if not isinstance(other, dict):
            try:
                other = dict(other)
            except TypeError:
                return id(self) < id(other)
        return dict(self) < other
        
    def __ge__(self, other):
        # TODO blech, change this
        if not isinstance(other, dict):
            try:
                other = dict(other)
            except TypeError:
                return id(self) >= id(other)
        return dict(self) >= other
        
    def __le__(self, other):
        # TODO blech, change this
        if not isinstance(other, dict):
            try:
                other = dict(other)
            except TypeError:
                return id(self) <= id(other)
        return dict(self) <= other

    def __len__(self):
        # keeping track of a separate length is currently considered to be
        # somewhat expensive per write and not necessarily always valuable
        # (that is, we don't always care what the length is). Therefore, we
        # have this quite expensive default approach. If you need a len often
        # and cheaply then you'll need to wrap the BForest and keep track of it
        # yourself.  This is expensive both because it wakes up every BTree and
        # bucket in the BForest, which might be quite a few objects; and
        # because of the merge necessary to create the composite tree (which
        # eliminates duplicates from the count).
        return len(self.tree())
    
    def setdefault(self, key, failobj=None):
        try:
            res = self[key]
        except KeyError:
            res = self[key] = failobj
        return res
    
    def pop(self, key, d=_marker):
        try:
            res = self[key]
        except KeyError:
            if d is self._marker:
                raise KeyError(key)
            else:
                return d
        else:
            del self[key]
            return res
    
    def popitem(self):
        key = self.minKey()
        val = self.pop(key)
        return key, val

    def __contains__(self, key):
        for b in self.buckets:
            if b.has_key(key):
                return True
        return False

    has_key = __contains__

    def copy(self):
        # this makes an exact copy, including the individual state of each 
        # bucket.  If you want a dict, cast it to a dict, or if you want
        # another one of these but with all of the keys in the first bucket,
        # call obj.__class__(obj)
        copy = self.__class__(count=len(self.buckets))
        for i in range(len(self.buckets)):
            copy.buckets[i].update(self.buckets[i])
        return copy
    
    def clear(self):
        for b in self.buckets:
            b.clear()
    
    def __nonzero__(self):
        for b in self.buckets:
            if bool(b):
                return True
        return False
    
    def rotateBucket(self):
        buckets = self.buckets
        b = buckets.pop()
        b.clear()
        buckets.insert(0, b)

class IOBForest(AbstractBForest):
    _treemodule = BTrees.family32.IO

class OIBForest(AbstractBForest):
    _treemodule = BTrees.family32.OI

class IIBForest(AbstractBForest):
    _treemodule = BTrees.family32.II

class LOBForest(AbstractBForest):
    _treemodule = BTrees.family64.IO

class OLBForest(AbstractBForest):
    _treemodule = BTrees.family64.OI

class LLBForest(AbstractBForest):
    _treemodule = BTrees.family64.II

class OOBForest(AbstractBForest):
    _treemodule = BTrees.family32.OO



##############################################################################
#
# Copyright (c) 2004 Zope Corporation and Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""test bforests

$Id: tests.py 29018 2005-02-02 15:28:36Z poster $
"""

import unittest
import datetime

import BTrees
import zope.testing

import zope.bforest
import zope.bforest.periodic

def StringGenerator(src='abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'):
    "infinite-ish unique string generator"
    for el in src:
        yield el
    for pre in StringGenerator(src):
        for el in src:
            yield pre + el

def NumberGenerator(number=0, interval=1):
    "infinite-ish unique number generator"
    while 1:
        yield number
        number += interval


oldDatetime = datetime.datetime
class _datetime(oldDatetime):
    _now = None
    @classmethod
    def now(klass, tzinfo=None):
        if tzinfo is None:
            return klass._now.replace(tzinfo=None)
        else:
            return klass._now.astimezone(tzinfo)
    def astimezone(self, tzinfo):
        return _datetime(
            *super(_datetime,self).astimezone(tzinfo).__reduce__()[1])
    def replace(self, *args, **kwargs):
        return _datetime(
            *super(_datetime,self).replace(
                *args, **kwargs).__reduce__()[1])
    def __repr__(self):
        raw = super(_datetime, self).__repr__()
        return "datetime.datetime%s" % (
            raw[raw.index('('):],)
    def __reduce__(self):
        return (argh, super(_datetime, self).__reduce__()[1])
def setNow(dt):
    _datetime._now = _datetime(*dt.__reduce__()[1])
def argh(*args, **kwargs):
    return _datetime(*args, **kwargs)

def setUp(test):

    datetime.datetime = _datetime
    test.globs['setNow'] = setNow

def tearDown(test):
    datetime.datetime = oldDatetime # cleanup

def test_suite():
    suite = unittest.TestSuite()
    numgen = iter(NumberGenerator()).next
    longgen = iter(NumberGenerator(BTrees.family32.maxint)).next
    strgen = iter(StringGenerator()).next
    suite.addTest(
        zope.testing.doctest.DocFileSuite(
            'bforest.txt',
            globs={'BForest': zope.bforest.IOBForest, 
                   'KeyGenerator': numgen, 
                   'ValueGenerator': strgen}))
    suite.addTest(
        zope.testing.doctest.DocFileSuite(
            'bforest.txt',
            globs={'BForest': zope.bforest.OIBForest, 
                   'KeyGenerator': strgen, 
                   'ValueGenerator': numgen}))
    suite.addTest(
        zope.testing.doctest.DocFileSuite(
            'bforest.txt',
            globs={'BForest': zope.bforest.IIBForest, 
                   'KeyGenerator': numgen, 
                   'ValueGenerator': numgen}))
    suite.addTest(
        zope.testing.doctest.DocFileSuite(
            'bforest.txt',
            globs={'BForest': zope.bforest.LOBForest, 
                   'KeyGenerator': longgen, 
                   'ValueGenerator': strgen}))
    suite.addTest(
        zope.testing.doctest.DocFileSuite(
            'bforest.txt',
            globs={'BForest': zope.bforest.OLBForest, 
                   'KeyGenerator': strgen, 
                   'ValueGenerator': longgen}))
    suite.addTest(
        zope.testing.doctest.DocFileSuite(
            'bforest.txt',
            globs={'BForest': zope.bforest.LLBForest, 
                   'KeyGenerator': longgen, 
                   'ValueGenerator': longgen}))
    suite.addTest(
        zope.testing.doctest.DocFileSuite(
            'bforest.txt',
            globs={'BForest': zope.bforest.OOBForest, 
                   'KeyGenerator': strgen, 
                   'ValueGenerator': strgen}))
    suite.addTest(
        zope.testing.doctest.DocFileSuite(
            'periodic.txt',
            globs={'BForest': zope.bforest.periodic.IOBForest, 
                   'KeyGenerator': numgen, 
                   'ValueGenerator': strgen},
            setUp=setUp, tearDown=tearDown))
    suite.addTest(
        zope.testing.doctest.DocFileSuite(
            'periodic.txt', 
            globs={'BForest': zope.bforest.periodic.OIBForest, 
                   'KeyGenerator': strgen, 
                   'ValueGenerator': numgen},
            setUp=setUp, tearDown=tearDown))
    suite.addTest(
        zope.testing.doctest.DocFileSuite(
            'periodic.txt', 
            globs={'BForest': zope.bforest.periodic.IIBForest, 
                   'KeyGenerator': numgen, 
                   'ValueGenerator': numgen},
            setUp=setUp, tearDown=tearDown))
    suite.addTest(
        zope.testing.doctest.DocFileSuite(
            'periodic.txt',
            globs={'BForest': zope.bforest.periodic.LOBForest, 
                   'KeyGenerator': longgen, 
                   'ValueGenerator': strgen},
            setUp=setUp, tearDown=tearDown))
    suite.addTest(
        zope.testing.doctest.DocFileSuite(
            'periodic.txt', 
            globs={'BForest': zope.bforest.periodic.OLBForest, 
                   'KeyGenerator': strgen, 
                   'ValueGenerator': longgen},
            setUp=setUp, tearDown=tearDown))
    suite.addTest(
        zope.testing.doctest.DocFileSuite(
            'periodic.txt', 
            globs={'BForest': zope.bforest.periodic.LLBForest, 
                   'KeyGenerator': longgen, 
                   'ValueGenerator': longgen},
            setUp=setUp, tearDown=tearDown))
    suite.addTest(
        zope.testing.doctest.DocFileSuite(
            'periodic.txt', 
            globs={'BForest': zope.bforest.periodic.OOBForest, 
                   'KeyGenerator': strgen, 
                   'ValueGenerator': strgen},
            setUp=setUp, tearDown=tearDown))
    return suite

if __name__ == '__main__': 
    unittest.main(defaultTest='test_suite')
